import math as m
from PIL import Image, ImageOps
import numpy as np
def normalize(ch_input):
    """
    normalizes an input array according to the rgb linearlization algorithm
    Args:
        ch_input: a 2d array containing pixel values of one rgb channel

    Returns: 
        void

    Dependencies: 
        requires 'pow' module from library 'math' to linearlize bright pixels
    """
    for i in range(0, len(ch_input)):
        for j in range(0, len(ch_input[0])):
            if(ch_input[i][j] <= 0.0405):
                ch_input[i][j] = ch_input[i][j] / 12.92
            else:
                ch_input[i][j] = m.pow( ((ch_input[i][j] + 0.055)/1.055), 2.4 )  

CHANNEL_WEIGHTS = [0.2126, 0.7152, 0.0722]
def rgb_to_grayscale (img_array):  
    """
    converts an array of rgb pixel values into grayscale using the grayscale conversion algorithm
    Args:
        img_array: a 3 dimensional array containing data on the image
                   more specifically, it is a numpy array obtained from calling
                   numpy.asarray on PIL.Image.open, and NOT an appended list of
                   rgb pixel values
    Returns: 
        gray_values: a 2 dimensional array containing grayscale converted pixel values for the image

    Dependencies: 
        reqires module 'Image' from library 'PIL' for valid input data
        requires library 'numpy' for valid input type
    """
    ch_data = [0] * 3
    for i in range(0, len(ch_data)):
        ch_data[i] = img_array[:,:,i] / 255
        normalize(ch_data[i])
        ch_data[i] = ch_data[i] * CHANNEL_WEIGHTS[i]

    gray_values = np.zeros((len(ch_data[0]), len(ch_data[0][0])))

    for i in range(0, len(gray_values)):
        for j in range(0, len(gray_values[0])):
            for k in range(0, len(ch_data)):
                gray_values[i][j] += ch_data[k][i][j] * 255 

    return gray_values

class image:
    in_path: str
    img_data: list
    ch_data: list

    zero_width: int
    zero_height: int

    TGT_WIDTH = 100
    TGT_HEIGHT = 100
    size: tuple
    set_target_size_called = False

    def set_in_path(self, path:str):
        """
        sets the path for a new image
        Args:
            path <string>:  the path to the image
        Returns: 
            void
        Dependencies: 
            None
        """
        self.in_path = path

    def set_tar_size(self, width, height):
        """
        sets the output target size to map the image to
        Args:
            width <int>: the desired output width
            height <int>: the desired output height
        Returns: 
            void
        Dependencies: 
            None
        """
        self.TGT_WIDTH = width
        self.TGT_HEIGHT = height
        self.set_target_size_called = True

    def new_image(self):
        """
        initializes a new image class. a path must be set but target size will default to 100 with no call
        Args:
            none
        Returns: 
            void
        Dependencies: 
            None
        """

        t = Image.open(self.in_path)
        (self.zero_width,self.zero_height) = (t.width // 2, t.height // 2)
        if(self.set_target_size_called == False):
            self.TGT_WIDTH = self.zero_width
            self.TGT_HEIGHT = self.zero_height

        self.size = (self.TGT_WIDTH, self.TGT_HEIGHT)
        t2 = ImageOps.pad(t, self.size, color = '#000')
        self.img_data = np.asarray(t2, dtype= np.uint8)
        self.img_data = self.img_data.copy() 

        
    
    def get_rgb_data(self):
        for i in range(0, 3):
            self.ch_data = np.zeros( (3, len(self.img_data), len(self.img_data[0]) ) )
            self.ch_data[i] = self.img_data[:,:,i] / 255
            normalize(self.ch_data[i])
            self.img_data[:,:,i] = self.ch_data[i] * 255

    
    def get_grayscale_data(self):
        self.ch_data = self.img_data / 255
        normalize(self.ch_data)
        self.img_data = self.ch_data * 255
    
    def show_image(self):
        img_out = Image.fromarray(self.img_data)
        img_out.show()
    


import math as m
import numpy as np
from Team19CustomUtils import generalUtils

def calculate_sigmoid(z: any):
    """
    calculates the position of an input z according to a rational normalization function
    args:
        z (float): any input number
    returns:
        output (float): the value of the function 1/ (1 + e^-x). A small deviation is added to this value to prevent e^-x from inflating to infinity or diminishing to 0

    """
    e0 = 0.00000000465661287 #2^-31, a value small enough that it should not affect calculation results while staying within most common bounds of negative power floats
                             #it additionally has the benefit of being albe to be precisely represented as a power of 2, greatly reducing floating point error
                             #this value has been entered manually to avoid fallbacking on outside modules, which may introduce their own inconsistencies when constructing stabilizer values
    
    output = 1/(1+ m.exp((-1 * z) + e0))
    #e0 stabilizes the denominator and prevents division by 0
    
    return output

def predict_proba(X: list, w: list, b: float):
    """
    calculates the probability for all examples in a dataset
    args:
        X (2d arr): feature matrix in which len(X) returns the number of examples and len(X[0]) returns the number of features
        w (1d arr): weight vector where len(w) matches with len(X[0])
        b (float): bias factor
    returns:
        z (1d arr): probability array where len(z) matches with len(x)

                        matrix multiplication between two matrices x * y and y * z will return an output matrix of x * z
                        in this case, matrix multiplication is performed between one of shape m * n and one of shape n * 1,
                        resulting in an output shape of m * 1

                        this value is then normalized to the sigmoid function to complete the probability array
    """
    z = X @ w + b
    for i in range(0, len(z)):
        z[i] = calculate_sigmoid(z[i])


    print(np.shape(z))
    return z

def predict_labels(X: list, w: list, b: float, threshold = 0.5):
    """
    converts probabilities to binary class labels (0 or 1)
    args:
        X (2d arr): feature matrix in which len(X) returns the number of examples and len(X[0]) returns the number of features
        w (1d arr): weight vector where len(w) matches with len(X[0])
        b (float): bias factor
        threshold (float): pre-defined comparison threshold. values over this will be designated 1 and values under this will be designated 0
                           the break-even point is equivalent to threshold, so values exactly equivalent to this value will additionally be designated 1

    returns:
        z(1d array): binary classified probability array where len(z) matches with len(X)

                        matrix multiplication between two matrices x * y and y * z will return an output matrix of x * z
                        in this case, matrix multiplication is performed between one of shape m * n and one of shape n * 1,
                        resulting in an output shape of m * 1
    """
    z = predict_proba(X, w, b)
    for i in range(0, len(z)):
        if(z[i] < threshold): 
            z[i] = 0
        else:
            z[i] = 1
    
    return z

def compute_loss_and_grads(X: list, y_true: list, w: list, b: float):
    """
    calculates and applies gradient loss to input data
    args:
        X (2d arr): feature matrix in which len(X) returns the number of examples and len(X[0]) returns the number of features
        y_true (1d arr): true label values
        w (1d arr): weight vector where len(w) matches with len(X[0])
        b (float): bias factor
    returns:
        loss (float): a loss value indicating the inaccuracy of the model
        dw (1d arr): step distance of weight vectors
        db (float): step distance of bias
    """
    y_pred = predict_labels(X, w, b)
    loss = get_loss(y_true, y_pred)
    
    temp1, temp2, dw, db = gradient_loss(X, w, b, y_true)

    return loss, dw, db

def get_loss(y_true: list, y_pred: list):
    """
    calculates the loss by comparing true y to predicted y
    args:
        y_true (1d arr): true label of input values
        y_pred (1d arr): predicted labels of values
    returns:
        loss (float): multi-vector loss sum for the entire dataset
    """
    m = len(y_true)
    loss = -1 * 1/m
    sum = 0
    for i in range(0, m):
        sum += y_true[i] * m.log(y_pred[i], 10) + (1 - y_true[i]) * m.log((1-y_pred[i]), 10)

    loss += sum
    return loss

def gradient_loss(X: list, w: list, b: float, y_true: list):
    """
    updates the values of w and b according to gradient loss function
    args:
        X (2d arr): feature matrix in which len(X) returns the number of examples and len(X[0]) returns the number of features
        w (1d arr): weight vector array
        b (float): bias factor
        y_true (1d arr): true label of input values
    returns:
        w (1d arr): adjusted weight vector array
        b (float): adjusted bias factor
        dL_w (1d arr): step distance of weight vector array
        dL_b (float): step distance of bias factor

    """


    y_pred = predict_labels(X, w, b)
    a = 0.001
    X_t = generalUtils.transpose_matrix(X)
    m = len(X)
    dL_w = 1/m * X_t @ (y_pred - y_true)
    dL_b = 1/m 
    dsum = 0
    for i in range(0, len(y_true)):
        dsum += (y_pred[i] - y_true[i])
    dL_b *= dsum
    
    w -= (a * dL_w)
    b -= (a * dL_b)

    return w, b, dL_w, dL_b

def train_logistical_regression(X_train: list, y_train: list, a: float, num_iters: int):
    """
    trains the model using gradient descent
    args:
        X_train (2d arr): training feature matrix
        y_train (1d arr): training labels array
        a (float): learning rate
        num_iters: upper limit to update cycles
    returns:
        w(1d arr): finalized weight data
        b (float): finalized bias data
        loss_history (1d arr): record of previous loss values before final iteration
    """
    m, n = X_train.shape

    w = np.random.randn(n) * 0.01
    b = 0.0


    loss_history = [0] * num_iters
    for i in range(0, num_iters):
        loss, dw, db = compute_loss_and_grads(X_train, y_train, w, b)
        w -= a * dw
        b -= b * dw
        loss_history[i] = loss

    return w, b, loss_history

def calculate_metrics(y_pred: list, y_true: list):
    """
    calcultes the efficacy in which the prediction labels match true labels
    args:
        y_pred (1d arr): predicted labels array
        y_true (1d arr): true labels array
    returns:
        accuracy (float): the accuracy rate of predictions that match true values
        error_rate (float): the inverse of accuracy - a measurement of how many predictions deviate from true values
    """
    accuracy = np.mean(y_pred == y_true)
    error = 1 - accuracy

    return accuracy, error

def evaluate_logistical_regression(X: list, y: list, w: list, b: float):
    """
    evaluates the effectiveness of a trained dataset on new data
    args:
        X (2d arr): sample data to evaluate upon
        y (1d arr): sample labels to evaluate upon
        w (1d arr): finalized weight vectors
        b (float): finalized bias scalar
    returns:
        accuracy (float): a measurement of the amount of successful predictions
        error_rate (float): a measurement of the amount of unsuccessful predictions
    """
    y_pred = predict_labels(X, w, b)
    accuracy, error_rate = calculate_metrics(y_pred, y)

    return accuracy, error_rate




import numpy as np

def get_ratio(num1, num2):
    """
    calculates the ratio between two numbers
    Args:
        num1: first number
        num2: second number
    Returns: 
        _: the ratio between the greater and lesser number (lesser/greater)
        state: a variable indicating which number is larger

    Dependencies: 
        None
    """

    if(num1 > num2):
        state = 1
        return ((num2/num1), state)

    elif(num2 > num1):
        state = 2
        return ( (num1/num2), state)

    else:
        return 0

def a_sum(input):
    """
    calculates the sum of an array
    Args:
        input: a one-dimensional array
    Returns: 
        sum: the sum of all values in the array

    Dependencies: 
        None
    """
    sum = 0
    for i in range (0, len(input)):
            sum += input[i]

    return sum

def a_mean(input):
    """
    calculates the mean of an array
    Args:
        input: a one-dimensional array
    Returns: 
        the mean of the one-dimensional array

    Dependencies: 
        None
    """
    return ( a_sum(input) / len(input) )

def transpose_matrix(input: list):
    """
    transposes an input matrix
    args:
        input (2d arr): input matrix which will be transposed
    returns:
        m_transposed (2d arr): transposition of input matrix
    """
    m_transposed = input.copy()
    for i in range(0, len(input)):
        for j in range(0, len(input[0])):
            m_transposed[i][j] = input[j][i]
    
    return m_transposed

def a_reduceDims(input:list):
    output = [0] * ( len(input) * len(input[0]) )
    k = 0
    for i in range(0, len(input)):
        for j in range(0, len(input[0])):
            output[k] = input[i][j]
            k += 1
    
    return output
            
def a_snip(input:list):
    l = 0
    for i in range(0, len(input)):
        if(input[i] != 0):
            l += 1

    output = [0] * l
    for i in range(0, output):
        output[i] = input[i]

    return output


def contains(input:list, target):
    """
    scans a list for presence of a target
    args:
        input (list): the list to check for the target
        target (dynamic): the target that the algorithm is checking for

    returns:
        1 - input list contains target
        0 - input list does not contain target
    """
    dims = np.ndim(input)
    if dims > 1:
        for i in range(dims, 1, - 1):
            input = a_reduceDims(input)
    
    for i in range(0, len(input)):
        if(input[i] == target):
            return 1
        
    return 0

def scanNumericalInput(mode:str, outputContainer, ranges:list, message: str):
    """
    scans for a number input (int or float)
    args:
        mode (string): 'range' - scans for valid inputs within specified range | 'match' - scans for valid inputs matching specified values
        outputContainer (dynamic): variable to store input. this variable's type is also used to predefine the expected type
        ranges (1d list): valid inputs 
        message (string): the prompt to display to the user

    returns:
        outputContainer (dynamic): valid user input
    """

    outputContainer = getI(outputContainer, message)
    match mode:
        case 'range':
            minVal = min(ranges)
            maxVal = max(ranges)

            while (outputContainer < minVal) or (outputContainer > maxVal):
                print('Invalid input range, please try again.')
                print('Valid inputs range from', minVal, 'to', maxVal )


                outputContainer = getI(outputContainer, message)
        case 'match': 
            while contains(ranges, outputContainer) == 0:
                
                print('Invalid input range, please try again.')


                outputContainer = getI(outputContainer, message)
        
        case _:
            return ValueError("unspecified or invalid mode declaration")
    return outputContainer
            
            


    
        
def getI(outputContainer, message: str):
#subfunction of scanNumericalInput
    while True:
        print(message, end = '')
        try:
            if isinstance(outputContainer, int) == True:
                outputContainer = int(input(''))
            elif isinstance(outputContainer, float) == True:
                outputContainer = float(input(''))
            else:
                return TypeError('Input container is not of int or float type')
            break

        except ValueError:
            print("Invalid input type, please try again")


    return outputContainer


    
        
    


import numpy as np


class tree:
    #the goal of procedural generation is to conserve computer resources and generate the tree only as required.
    #due to the unique constraints of the situation, that being the user is not able to perform backwards manipulation,
    #preceding portions of the tree can be "forgotten" and future paths are generated as-needed
    #using computer memory for this task makes for a simple solution as the code already operates upon memory by default

    #this allows for the loading of extremely high-dimension matrices with unknown projected memory sizes to be
    #represented with a predetermined amount of memory space using lower-dimension local representations
    #specifically, the parts of the larger matrix loaded composes of a "present" buffer and a "look-ahead buffer"

    #this is most often used in games featuring linear progression (ie. the player can only move forward to the next map, level, etc.)
    #this is also another reason why most of these games do not feature extensive save capabilities and instead opt for gameplay loops 
    #as any saved data would probably be invalid due to a limited amount of memory being alternated to compress data


    seed = [0, 0, 0] 
    currentDepth = 1 #a variable used for external monitoring of current 'explored' depth
    presentBuffer = None 
    lookAheadBuffer = None
    populator = None
    seeded = 0
    
    

    def setSeed(self, depth:int, branchPopulation:int, divergence:int, populator:list, random: bool) -> None:
        """
        scans for a number input (int or float)
        args:
            depth (int): specified depth in which procedural generation will stop
            branchPopulation (int): the list population on each node
            divergence (int): the amount of future nodes that generate from the current node
            populator (list): container of values to fill each node
            random (bool): toggles random mode (uses user specifications as a ceiling for random generation instead of fixed values)

        returns:
            ValueError if 'random' argument has not been assigned correctly
            void otherwise

    """
        if random == False:
            self.seed[0] = depth
            self.seed[1] = branchPopulation
            self.seed[2] = divergence
            self.presentBuffer = [0] * self.seed[1]
            self.lookAheadBuffer = np.zeros( (self.seed[2], self.seed[1] ), dtype=int)
            self.populator = populator

            self.seeded = 1
        elif random == True:
            self.seed[0] = np.random.randint(1, depth)
            self.seed[1] = np.random.randint(1, branchPopulation)
            self.seed[2] = np.random.randint(1, divergence)
            self.presentBuffer = [0] * self.seed[1]
            self.lookAheadBuffer = np.zeros( (self.seed[2], self.seed[1] ), dtype=int)
            self.populator = populator

            self.seeded = 1
        else:
            return ValueError("Invalid assignment of 'random' argument ")

    def populateBuffers(self):
        match self.seeded:
            case 1:
                for i in range(0, len(self.presentBuffer)):
                    #fills the present buffer with random numbers from the populator list
                    self.presentBuffer[i] = self.populator[np.random.randint(0, len(self.populator))]
                self.populateFutureBuffer()
                return None
            case _:
                return RuntimeError("the specified tree is currently unseeded")


    def advanceDepths(self, selector:int):
        match self.seeded:
            case 1:
                if(selector > len(self.lookAheadBuffer)):
                    return OverflowError("selector %d ", selector ,"exceeds possible selections %d", len(self.lookAheadBuffer))
                else:
                    #set the selected index of future buffer to current buffer
                    self.currentDepth += 1
                    self.presentBuffer = self.lookAheadBuffer[selector - 1]
                    #regenerate future buffer
                    self.populateFutureBuffer()
                    return None

                    #this operation effectively 'forgets' the previous buffer set
            case _:
                return RuntimeError("the specified tree is currently unseeded")

    def populateFutureBuffer(self):
        match self.seeded:
            case 1:
                for i in range(0, len(self.lookAheadBuffer)):
                    #populates future buffer with different 'choices' of random values
                    #each subset in the future buffer is of equivalent length to the current buffer
                    for j in range(0,len(self.lookAheadBuffer[0])):
                        self.lookAheadBuffer[i][j] = self.populator[np.random.randint(0,len(self.populator))]
                return None
            case _:
                return RuntimeError("the specified tree is currently unseeded")
        






    
        

    

            

